/*-
 * Copyright (C)2005 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: insert.c,v 1.9 2014/06/14 12:00:34 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "multilar.h"
#include "mular_int.h"

#define LOCK(A)
#define UNLOCK(A)
#define SHIFT(A, B, C) mular_walk(mular, (A), (B), (C), stderr, NULL, mular_plus, NULL, NULL, NULL)
#ifdef __amd64__
#define PTOFI (u_int32_t)(u_int64_t)
#else
#define PTOFI (u_int32_t)
#endif

#ifdef DEBUG
static mular_indir *ctrl[DEBUG] =
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

void
mular_place(mular_indir *p, int n) {
    int i, j;
    for (j = 0; j < n; j++) {
        for (i = 0; i < DEBUG; i++) {
            if  (&p[j] == ctrl[i]) break;
            if  (!ctrl[i]) {
                ctrl[i] = &p[j];
                break;
}   }   }   }
#endif /* DEBUG */

static void
mular_plus(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, int ex, FILE *out) {
    ifBLIN_FLAG = mular->flags;

    idx->index++;
    ifBLIN_QO2
        fprintf(stderr, " plus %08X(%u) level=%u\n", PTOFI idx, (u_int)idx->index, level);
}

static size_t
mular_todiv(mular_descriptor *mular, size_t relidx, int level) {
    size_t todiv;
    ssize_t delta;

    todiv = (mular->flags & MULAR_HALF) ? SIZE(level) / 2 : relidx;
    if  (!(mular->flags & MULAR_STRI)) {
        if  (SIZE(level) / 2 < relidx) {
            delta = relidx - SIZE(level) / 2;
            delta = -delta / 4;
        } else {
            delta = (SIZE(level) / 2 - relidx) / 4;
        }
        todiv += (mular->flags & MULAR_HALF) ? -delta : delta;
    }
    return(todiv);
}

static void *
mular_div(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir dubl[2], int level) {
    ifBLIN_FLAG = mular->flags;
    size_t todiv, relidx;
    void *res;

    ifBLIN_QO3
        fprintf( stderr, " +div %u(%u..%u) [%08X %08X] level=%u\n"
               , (u_int)indx, (u_int)mindx, (u_int)dubl[0].index, PTOFI&dubl[0], PTOFI&dubl[1], level
               );
    if  (indx < mindx || indx > dubl[0].index || level < 1) {
        errno = EINVAL;
        ifBLIN_QW1( " div: index=%u(%u..%u) level=%u size=%u"
                  , (u_int)indx, (u_int)mindx, (u_int)dubl[0].index
                  , level, (u_int)SIZE(level)
                  );
        res = NULL;
    } else if (!(TERM(dubl[1]) =  mular->flags & MULAR_ZERO
                                ? calloc(SIZE(level), level == 1 ? mular->unit : sizeof(mular_indir))
                                : malloc(level == 1 ? LTERM(SIZE(level)) : LNEXT(SIZE(level)))
              ) ) {
        errno = ENOMEM;
        ifBLIN_QW0(" div");
        res = NULL;
    } else if (level == 1) {

        relidx = indx - mindx;
        todiv = mular_todiv(mular, relidx, level);
        ifBLIN_QO4 {
            mular_displine( mular, relidx, mindx, dubl, level, stderr
                          , MULAR_DISPLINE_D | MULAR_DISPLINE_P | (MULAR_DISPLINE_N & todiv)
                          );
        }
        if  (mindx + todiv > dubl[0].index) todiv = dubl[0].index - mindx;
        dubl[1].index = dubl[0].index + 1;

        /* todiv     ,  relidx,         *
         *          *
         *    relidx                 */
        if  (relidx < todiv) {
            ifBLIN_QO2
                fprintf(stderr, "#1 idx=%u, div=%u\n", (u_int)relidx, (u_int)todiv);
            res = TD(0, relidx);
            TCP(TD(1, 0), TD(0, todiv), SIZE(level) - todiv);
            TCP(TD(0, relidx + 1), res, todiv - relidx);
            ifMULAR_ZERO {
                TZR(TD(0, todiv + 1), SIZE(level) - todiv - 1);
                TZR(res, 1);
            }
            dubl[0].index = mindx + todiv + 1;
        } else if (relidx > todiv) {
            ifBLIN_QO2
                fprintf(stderr, "#2 idx=%u, div=%u\n", (u_int)relidx, (u_int)todiv);
            res = TD(1, relidx - todiv);
            TCP(TD(1, 0), TD(0, todiv), relidx - todiv);
            TCP(TD(1, relidx - todiv + 1), TD(0, relidx), SIZE(level) - relidx);
            ifMULAR_ZERO TZR(TD(0, todiv), SIZE(level) - todiv);
            dubl[0].index = mindx + todiv;
        } else if (todiv == 0) {                                                  /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "#3 idx=%u, div=%u\n", (u_int)relidx, (u_int)todiv);
            res = TD(0, 0);                                                 /* may be spap 0 and 1 ? */
            TCP(TD(1, 0), TD(0, 0), SIZE(level));
            ifMULAR_ZERO TZR(res, SIZE(level));
            dubl[0].index = mindx + 1;
        } else if (todiv == dubl[0].index) {                                      /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "#4 idx=%u, div=%u\n", (u_int)relidx, (u_int)todiv);
            res = TD(1, 0);
        } else {                                                                  /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "#5 idx=%u, div=%u\n", (u_int)relidx, (u_int)todiv);
            res = TD(1, 0);
            TCP(TD(1, 1), TD(0, todiv), SIZE(level) - todiv);
            ifMULAR_ZERO TZR(TD(0, todiv), SIZE(level) - todiv);
            dubl[0].index = mindx + todiv;
        }
        ifBLIN_QO2 mular_displine(mular, relidx, mindx, dubl, level, stderr, 2);
        ifBLIN_QO2 fprintf( stderr, "0=%08X[%u] 1=%08X[%u] r=%08X\n"
                          , PTOFI TERM(dubl[0]), (u_int)dubl[0].index
                          , PTOFI TERM(dubl[1]), (u_int)dubl[1].index, PTOFI res
                          );
    } else {
        ssize_t cont;
        mular_indir *t;
        size_t i, j;

        MULAR_INBUG(NEXT(dubl[1]), 1);
        for (i = 0, j = mindx; ((cont = ICONN(dubl, i, t)) >= 0); i++) {
            ifBLIN_QO4
                fprintf( stderr, " div i=%u j=%u cont=%d t->index=%u indx=%u t=%08X\n"
                       , (u_int)i, (u_int)j, (int)cont, (u_int)t->index, (u_int)indx, PTOFI t
                       );
            if  (t->index > indx) break;
            if  (!cont) break;
            j = t->index;
        }
        relidx = i++;
        todiv = mular_todiv(mular, relidx, level);
        dubl[1].index = dubl[0].index + 1;
        ifBLIN_QO4 {
            mular_displine( mular, relidx, mindx, dubl, level, stderr
                          , MULAR_DISPLINE_D | MULAR_DISPLINE_P | (MULAR_DISPLINE_N & todiv)
                          );
        }
        for (; cont && ((cont = ICONN(dubl, i, t)) >= 0);) {
            ifBLIN_QO3 fprintf(stderr, " div i=%u cont=%d t=%08X\n", (u_int)i, (int)cont, PTOFI t);
            SHIFT(mindx, t, level - 1);
            i++;
            if  (!cont) break;
        }

        /* todiv     ,  relidx,         *
         *          *
         *    relidx                 *
         *  , ,   relidx            *
         *                                       */
        if  (relidx < todiv) {
            ifBLIN_QO2
                fprintf(stderr, "+1 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(0, relidx);
            NCP(ND(1, 0), ND(0, todiv), i - todiv);
            NCP(ND(0, relidx + 2), ND(0, relidx + 1), todiv - relidx - 1);
            ifMULAR_ZERO {
                NZR(ND(0, todiv + 1), SIZE(level) - todiv - 1);
                NZR(ND(0, relidx + 1), 1);
            }
            res = mular_div(mular, indx, j, t, level - 1);
            dubl[0].index = ND(0, todiv)->index;
        } else if (relidx > todiv) {
            ifBLIN_QO2
                fprintf(stderr, "+2 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(1, relidx - todiv);
            NCP(ND(1, 0), ND(0, todiv), relidx - todiv + 1);
            NCP(ND(1, relidx - todiv + 2), ND(0, relidx + 1), i - relidx - 1);
            ifMULAR_ZERO NZR(ND(0, todiv), SIZE(level) - todiv);
            res = mular_div(mular, indx, j, t, level - 1);
            dubl[0].index = ND(0, todiv - 1)->index;
        } else if (todiv == 0) {                                                  /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "+3 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(0, 0);
            dubl[0].index = ND(0, 0)->index + 1;
            NCP(ND(1, 0), ND(0, 1), i - 1);
            ifMULAR_ZERO NZR(ND(0, 1), SIZE(level) - 1);
            res = mular_div(mular, indx, j, t, level - 1);
        } else if (todiv == i - 1) {                                              /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "+4 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(1, 0);
            dubl[0].index = ND(0, todiv - 1)->index;
            NCP(ND(1, 0), ND(0, todiv), 1);
            ifMULAR_ZERO NZR(ND(0, todiv), 1);
            res = mular_div(mular, indx, j, t, level - 1);
        } else if (mular->flags & MULAR_UPPE) {                                   /* relidx == todiv */
            ifBLIN_QO2
                fprintf(stderr, "+5 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(1, 0);
            dubl[0].index = ND(0, todiv - 1)->index;
            NCP(ND(1, 0), ND(0, todiv), 1);
            NCP(ND(1, 2), ND(0, todiv + 1), i - todiv - 1);
            ifMULAR_ZERO {
                NZR(ND(0, todiv), SIZE(level) - todiv);
                NZR(ND(1, 1), 1);
            }
            res = mular_div(mular, indx, j, t, level - 1);
        } else {
            ifBLIN_QO2
                fprintf(stderr, "+6 reli=%u div=%u i=%u\n", (u_int)relidx, (u_int)todiv, (u_int)i);
            t = ND(0, todiv);
            dubl[0].index = t->index + 1;
            NCP(ND(1, 0), ND(0, todiv + 1), i - todiv - 1);
            ifMULAR_ZERO NZR(ND(0, todiv + 1), SIZE(level) - todiv - 1);
            res = mular_div(mular, indx, j, t, level - 1);
        }
        ifBLIN_QO4 mular_displine(mular, relidx, mindx, dubl, level, stderr, 2);
    }
    ifBLIN_QO3 fprintf(stderr, " -div %08X\n", PTOFI res);
    return(res);
}

static void *
mular_dub(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level) {
    ifBLIN_FLAG = mular->flags;
    void *res;

    ifBLIN_QO3 fprintf( stderr, " +dub flags=%08X i=%u[%u..%u]@%08X level=%d\n"
                      , mular->flags, (u_int)indx, (u_int)mindx, (u_int)idx->index
                      , PTOFI idx, level
                      );
    if  (indx < mindx || indx > idx->index) {
        errno = 0;
        ifBLIN_QO1 fprintf( stderr, " mular_dub: index=%u(%u..%u) level=%u size=%u\n"
                          , (u_int)indx, (u_int)mindx, (u_int)idx->index, level
                          , (u_int)SIZE(level)
                          );
        res = NULL;
    } else if (level > mular->level || level < 2) {
        errno = EINVAL;
        ifBLIN_QO0 fprintf(stderr, "level too big: %u > %u\n", level, mular->level);
        res = NULL;
    } else {
        ssize_t cont;
        mular_indir *t, *f;
        size_t i, j, relidx, midx;

        for (i = 0; i < SIZE(level) && ICONT(idx, i) > 0; i++) {}
        if  (++i < SIZE(level)) {
            for (i = 0, j = mindx; ((cont = ICONN(idx, i, t)) >= 0); i++) {
                if  (t->index > indx) break;
                if  (!cont) break;
                j = t->index;
            }
            relidx = i++;
            midx = j;
            ifBLIN_QO4 {
                mular_displine( mular, relidx, midx, idx, level, stderr
                              , MULAR_DISPLINE_P | 1
                              );
                mular_displine( mular, relidx, midx, t, level - 1, stderr
                              , MULAR_DISPLINE_P | 1
                              );
            }
            for (; cont && ((cont = ICONN(idx, i, f)) >= 0);) {
                SHIFT(j, f, level - 1);
                i++;
                if  (!cont) break;
                j = f->index;
            }
            NCP(&t[2], &t[1], i - relidx - 1);
            ifMULAR_ZERO NZR(&t[1], 1);
            res = mular_div(mular, indx, midx, t, level - 1);
            idx->index++;
        } else {
            errno = EINVAL;
            ifBLIN_QO0
                fprintf( stderr, "No space in table: %u <= %u level=%d for indx=%u\n"
                       , (u_int)SIZE(level), (u_int)i, level, (u_int)indx
                       );
            res = NULL;
    }   }
    ifBLIN_QO3 fprintf( stderr, " -dub i=%u[%u..%u] level=%d res=%08X\n"
                      , (u_int)indx, (u_int)mindx, (u_int)idx->index, level, PTOFI res
                      );
    return(res);
}

#include "states.h"

static void *
mular_get(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level) {
    ifBLIN_FLAG = mular->flags;
    void *res;

    ifBLIN_QO3 fprintf( stderr, " +get index=%u(%u..%u) level=%u size=%u\n"
                      , (u_int)indx, (u_int)mindx, (u_int)idx->index
                      , level, (u_int)SIZE(level)
                      );
    errno = 0;
    if  (indx < mindx || indx > idx->index || level < 1) {
        errno = EINVAL;
        ifBLIN_QW0( " mular_get: index=%u(%u..%u) level=%u size=%u"
                  , (u_int)indx, (u_int)mindx, (u_int)idx->index
                  , level, (u_int)SIZE(level)
                  );
        res = NULL;
    } else if (level == 1) {
        if  (idx->index >= mindx + SIZE(level)) {
            ifBLIN_QO1
                fprintf( stderr, " mular_get: level=%u size=%u idx->index-mindx=%u-%u\n"
                       , level, (u_int)SIZE(level), (u_int)idx->index, (u_int)mindx
                       );
            res = NULL;
        } else {
            res = ATERM(*idx, indx - mindx);
            if  (idx->index != indx) {
                TCP(ATERM(*idx, indx - mindx + 1), res, idx->index - indx);
                ifMULAR_ZERO TZR(res, 1);
            }
            idx->index++;
        }
    } else {
        size_t segi, segmin, nextmin = 0;
        mular_indir *segment = NULL;
        u_int32_t control, class, state;

        for ( segi = 0, segmin = mindx, res = NULL, control = 0, state = d0
            ; state < dx
            ; state = control & C_state
            ) {
            class = 0;
            if  (res) class |= Sr;
            if  (segmin <= indx) class |= Si;
            if  (segi < SIZE(level)) class |= Sx;
            control = automa[class][state];
            ifBLIN_QO3
                fprintf( stderr, ".%08X %08X i=%u<%u(%d) res=%08X (min=%u)<=(indx=%u)\n"
                       , class, control, (u_int)segi, (u_int)SIZE(level)
                       , level, PTOFI res, (u_int)segmin, (u_int)indx
                       );
            if  (control & Cp) {
                segment = &NEXT(*idx)[segi];
                nextmin = segment->index;
                if  (segment->index == idx->index) class |= Se;
                if  (segment->index < idx->index) class |= Ss;
                if  (segment->index >= indx) class |= Sa;
                control = automa[class][state];
                ifBLIN_QO3
                    fprintf( stderr, "@%08X %08X %u<=%08X->%u< =%u\n"
                           , class, control, (u_int)indx, PTOFI segment, (u_int)segment->index
                           , (u_int)idx->index
                           );
            }
            if  (control & Cr) res = mular_get(mular, indx, segmin, segment, level - 1);
            if  (control & Cs) SHIFT(mindx, segment, level - 1);
            if  (control & Cd) res = mular_dub(mular, indx, mindx, idx, level);
            if  (control & Cy) idx->index++;
            if  (control & Ci) {
                segmin = nextmin;
                segi++;
            }
            if  (control & Ce) {
                ifBLIN_QO3
                    fprintf( stderr, " mular_get: Internal error class=%08X control=%08X\n"
                           , class, control
                           );
    }   }   }
    ifBLIN_QO3 fprintf(stderr, " -get %08X\n", PTOFI res);
    return(res);
}

void *
mular_insert(mular_descriptor *mular, size_t indx) {
    ifBLIN_FLAG = mular->flags;
    void *res;

    ifBLIN_QO2 fprintf( stderr, "+mular_insert flags=%08X i=%u up=%u level=%d\n"
                      , mular->flags, (u_int)indx, (u_int)mular->top.index, mular->level
                      );
    LOCK(mular);
    errno = 0;
    res = NULL;
    if  (mular->flags & MULAR_FIXD) {
        errno = EINVAL;
        ifBLIN_QW0("mular_insert: mular fixed");
    } else if  (indx > mular->top.index) {
        errno = EINVAL;
        ifBLIN_QW0("mular_insert: void array, index=%u", (u_int)indx);
    } else if (mular->level == 0) {
        res = mular->flags & MULAR_ZERO ? calloc(SIZE(1), mular->unit) : malloc(LTERM(SIZE(1)));
        if  (!res) {
            errno = ENOMEM;
            ifBLIN_QW0("mular_insert #4");
        } else {
            TERM(mular->top) = res;
            mular->level = 1;
            mular->top.index = 1;
        }
    } else if (!(res = mular_get(mular, indx, 0, &mular->top, mular->level))) {
        size_t i;

        if  (mular->level == 1) {
            i = mular->top.index;
        } else {
            for (i = 0; i < SIZE(mular->level) && ICONT(&(mular->top), i) > 0; i++) {}
        }
        if  (++i >= SIZE(mular->level)) {
        ;   if  (mular->lemax > mular->level) {
                mular_indir *t;

        ;       if  (!(t =  mular->flags & MULAR_ZERO
                          ? calloc(SIZE(mular->level + 1), sizeof(mular_indir))
                          : malloc(LNEXT(SIZE(mular->level + 1)))
                    ) ) {
        ;           errno = ENOMEM;
        ;           ifBLIN_QW0("mular_insert #3");
        ;       } else {
        ;           MULAR_INBUG(t, 1);
        ;           NCP(t, &mular->top, 1);
        ;           NEXT(mular->top) = t;
        ;           mular->level++;
        ;           res = mular_dub(mular, indx, 0, &(mular->top), mular->level);
        ;       }
        ;   } else if  (mular->level == 1) {
        ;       if  (!(TERM(mular->top) = realloc(TERM(mular->top), LTERM(SIZE(mular->level)) * 2))) {
        ;           errno = ENOMEM;
        ;           ifBLIN_QW0("mular_insert #1");
        ;           res = NULL;
        ;       } else {
        ;           ifMULAR_ZERO TZR(ATERM(mular->top, SIZE(mular->level)), SIZE(mular->level));
        ;           SIZE(mular->level) *= 2;
        ;           res = mular_get(mular, indx, 0, &(mular->top), mular->level);
        ;       }
        ;   } else {
        ;       if  (!(NEXT(mular->top) = realloc(NEXT(mular->top), LNEXT(SIZE(mular->level)) * 2))) {
        ;           errno = ENOMEM;
        ;           ifBLIN_QW0("mular_insert #2");
        ;       } else {
        ;           MULAR_INBUG(NEXT(mular->top), 2);
        ;           ifMULAR_ZERO NZR(&NEXT(mular->top)[SIZE(mular->level)], SIZE(mular->level));
        ;           SIZE(mular->level) *= 2;
                    res = mular_dub(mular, indx, 0, &(mular->top), mular->level);
        }   }   }
        if  (!res) {
            errno = EINVAL;
            ifBLIN_QW0("mular_insert #5");
    }   }
    ifBLIN_QO2 fprintf( stderr, "-mular_insert up=%u level=%d res=%08X\n"
                      , (u_int)mular->top.index, mular->level, PTOFI res
                      );
    UNLOCK(mular);
    return(res);
}

void *
mular_add(mular_descriptor *mular) {
    return(mular_insert(mular, mular->top.index));
}
